import datetime
from types import NoneType, UnionType
from typing import Any, List, Union, assert_never, get_args, get_origin

from pydantic import BaseModel
from pydantic_core import PydanticUndefined

from adsdatahub.restapi.schemas.field_type import ArrayTypeDict, FieldTypeDict, TypeDict
from adsdatahub.restapi.schemas.parameter_type import ParameterTypeDict
from adsdatahub.restapi.schemas.parameter_value import (
    ArrayValueDict,
    ArrayValueValuesDict,
    ParameterValueDict,
    ValueDict,
)


def convert_pydantic_parameter_types(params: BaseModel) -> dict[str, ParameterTypeDict]:
    return {
        name: _convert_parameter_type(
            convert_pydantic_parameter_type(field_info.annotation),
            field_info.description,
            field_info.get_default(call_default_factory=True),
        )
        for name, field_info in params.model_fields.items()
    }


def _convert_parameter_type(
    field_type: FieldTypeDict,
    description: str | None,
    default_value: Any,
) -> ParameterTypeDict:
    parameter_type: ParameterTypeDict = {"type": field_type}

    if description is not None:
        parameter_type["description"] = description

    if default_value is not PydanticUndefined:
        parameter_type["defaultValue"] = convert_pydantic_parameter_value(default_value)

    return parameter_type


def convert_pydantic_parameter_values(
    params: BaseModel,
) -> dict[str, ParameterValueDict]:
    return {key: convert_pydantic_parameter_value(value) for key, value in params}


def convert_pydantic_parameter_type(field_type: type | None) -> FieldTypeDict:
    if field_type is None:
        raise TypeError("Field annotation cannot be None")

    origin_type = get_origin(field_type)
    if origin_type is None:
        if issubclass(field_type, bool):
            return TypeDict({"type": "BOOL"})

        elif issubclass(field_type, str):
            return TypeDict({"type": "STRING"})

        elif issubclass(field_type, int):
            return TypeDict({"type": "INT64"})

        elif issubclass(field_type, float):
            return TypeDict({"type": "FLOAT64"})

        elif issubclass(field_type, datetime.datetime):
            return TypeDict({"type": "TIMESTAMP"})

        elif issubclass(field_type, datetime.date):
            return TypeDict({"type": "DATE"})

    elif origin_type is list or origin_type is List:
        element_type = get_args(field_type)[0]
        return ArrayTypeDict(
            {"arrayType": convert_pydantic_parameter_type(element_type)}
        )

    elif origin_type is UnionType or origin_type is Union:
        element_types = get_args(field_type)
        for element_type in element_types:
            if element_type is NoneType:
                continue
            return convert_pydantic_parameter_type(element_type)

    raise TypeError(f"Unsupported type: {field_type}")


def convert_pydantic_parameter_value(value: Any) -> ParameterValueDict:
    if value is None:
        # NOTE: nullを渡す方法がわかっていません。
        #       値フィールドにnullを渡すと、次のようなエラーになることはわかっています。
        #
        #      ```json
        #      {"code": 3, "message": "Value for parameter \"value\" not given and no default value found. (Error 0410)"}
        #      ```
        #
        # See: https://developers.google.com/ads-data-hub/reference/rest/v1/FieldType?hl=ja

        raise NotImplementedError("null is not supported yet.")

    match value:
        case str():
            return ValueDict({"value": value})

        case bool():
            return ValueDict({"value": str(value).upper()})

        case int() | float():
            return ValueDict({"value": str(value)})

        case datetime.date() | datetime.datetime():
            return ValueDict({"value": value.isoformat()})

        case list():
            return ArrayValueDict(
                {
                    "arrayValue": ArrayValueValuesDict(
                        {"values": [convert_pydantic_parameter_value(v) for v in value]}
                    )
                }
            )

        case _ as unreachable:
            assert_never(unreachable)



import pandas as pd
from adsdatahub.client.query_job.query_job import QueryJob
from google.cloud import bigquery


class RealQueryJob(QueryJob):
    def __init__(self, job: bigquery.QueryJob) -> None:
        self._job = job

    def to_dataframe(self) -> pd.DataFrame:
        return self._job.to_dataframe()



import datetime
from typing import assert_never

from adsdatahub.restapi.schemas.field_type import (
    ArrayTypeDict,
    FieldTypeDict,
    TypeDict,
)
from adsdatahub.restapi.schemas.parameter_type import (
    ParameterTypeDict,
)
from adsdatahub.restapi.schemas.parameter_value import (
    ArrayValueDict,
    ArrayValueValuesDict,
    ParameterValueDict,
    ValueDict,
)

PrimitivePythonParameter = (
    str
    | int
    | float
    | bool
    | datetime.date
    | datetime.datetime
    | list["PrimitivePythonParameter"]
)


def convert_primitive_parameter_types(
    params: dict[str, PrimitivePythonParameter],
) -> dict[str, ParameterTypeDict]:
    return {
        key: ParameterTypeDict({"type": convert_primitive_parameter_type(value)})
        for key, value in params.items()
    }


def convert_primitive_parameter_values(
    params: dict[str, PrimitivePythonParameter],
) -> dict[str, ParameterValueDict]:
    return {
        key: convert_primitive_parameter_value(value) for key, value in params.items()
    }


def convert_primitive_parameter_type(
    value: PrimitivePythonParameter,
) -> FieldTypeDict:
    match value:
        case str():
            return TypeDict({"type": "STRING"})

        case int():
            return TypeDict({"type": "INT64"})

        case float():
            return TypeDict({"type": "FLOAT64"})

        case bool():
            return TypeDict({"type": "BOOL"})

        case datetime.date():
            return TypeDict({"type": "DATE"})

        case datetime.datetime():
            return TypeDict({"type": "TIMESTAMP"})

        case list():
            return ArrayTypeDict(
                {
                    "arrayType": convert_primitive_parameter_type(
                        # NOTE: 配列の要素が空の場合は、型を推測できないため、文字列を仮に入れる。
                        #       クエリ文の側でパラメータの型が明確である場合、文字列は自動で変換処理されるため、最も安全な選択肢となる。
                        value[0] if len(value) > 0 else "STRING"
                    )
                }
            )

        case _ as unreachable:
            assert_never(unreachable)


def convert_primitive_parameter_value(
    value: PrimitivePythonParameter,
) -> ParameterValueDict:
    match value:
        case str():
            return ValueDict({"value": value})

        case bool():
            return ValueDict({"value": str(value).upper()})

        case int() | float():
            return ValueDict({"value": str(value)})

        case datetime.date() | datetime.datetime():
            return ValueDict({"value": value.isoformat()})

        case list():
            return ArrayValueDict(
                {
                    "arrayValue": ArrayValueValuesDict(
                        {
                            "values": [
                                convert_primitive_parameter_value(v) for v in value
                            ]
                        }
                    )
                }
            )

        case _ as unreachable:
            assert_never(unreachable)



from pydantic import BaseModel

from adsdatahub.client.parameters.primitive_parameters import (
    PrimitivePythonParameter,
    convert_primitive_parameter_types,
    convert_primitive_parameter_values,
)
from adsdatahub.client.parameters.pydantic_parameters import (
    convert_pydantic_parameter_types,
    convert_pydantic_parameter_values,
)
from adsdatahub.restapi.schemas.parameter_type import ParameterTypeDict
from adsdatahub.restapi.schemas.parameter_value import ParameterValueDict

# TODO: 配列の型、デフォルト値などをサポートする必要があるが、後回し。
#
# TODO: デフォルト値や NULL のサポートをするためには、 dataclass や pydantic を利用する必要がある。
#
# NOTE: ドキュメントの記載は一貫性がない。
#       1. STRING または INT64 のプリミティブ型に加えて、配列と構造体をサポートしているように見える。
#          https://developers.google.com/ads-data-hub/reference/rest/v1/FieldType?hl=ja
#
#       2. AdsDataHub のコンソールからはさらに多くのパラメータの型を扱うことができるようだ。
#          https://developers.google.com/ads-data-hub/guides/run-queries?hl=ja#parameter_types
#
PythonParameter = PrimitivePythonParameter
"""
クエリのパラメータとして使える Python の型。
"""


def convert_parameter_types(
    parameter: dict[str, PrimitivePythonParameter] | BaseModel,
) -> dict[str, ParameterTypeDict]:
    if isinstance(parameter, BaseModel):
        return convert_pydantic_parameter_types(parameter)

    else:
        return convert_primitive_parameter_types(parameter)


def convert_parameter_values(
    parameter: dict[str, PrimitivePythonParameter] | BaseModel,
) -> dict[str, ParameterValueDict]:
    if isinstance(parameter, BaseModel):
        return convert_pydantic_parameter_values(parameter)

    else:
        return convert_primitive_parameter_values(parameter)



import pandas as pd
from adsdatahub.client.query_job.query_job import QueryJob


class MockQueryJob(QueryJob):
    def __init__(self, dataframe: pd.DataFrame):
        self._dataframe = dataframe

    def to_dataframe(self) -> pd.DataFrame:
        return self._dataframe



import datetime
from types import NoneType
from typing import Any, Self, TypeVar

from pydantic import BaseModel
from typing_extensions import override

import adsdatahub
from adsdatahub.client.customer.customer import CustomerClient
from adsdatahub.client.parameters import (
    PythonParameter,
)
from adsdatahub.client.query_result import QueryResult
from adsdatahub.exceptions import AdsDataHubMockDataTypeError, AdsDataHubMockMethodError
from adsdatahub.types import CustomerId

GenericResponse = TypeVar("GenericResponse")


class MockCustomerClient(CustomerClient):
    def __init__(
        self,
        client: adsdatahub.MockClient,
        customer_id: CustomerId,
    ):
        self._client = client
        self.customer_id = customer_id
        self._store: list[tuple[str, Any]] = []

    def inject_query_response(self, response: QueryResult | Exception) -> Self:
        self._store.append(("query", response))

        return self

    def inject_validate_response(self, response: None | Exception = None) -> Self:
        self._store.append(("validate", response))

        return self

    def _provide_response(
        self, expected_method: str, expected_response_type: type[GenericResponse]
    ) -> GenericResponse:
        method, response = self._store.pop(0)

        if method != expected_method:
            raise AdsDataHubMockMethodError(method, expected_method)

        if isinstance(response, Exception):
            raise response

        if isinstance(response, expected_response_type):
            raise AdsDataHubMockDataTypeError(type(response), expected_response_type)

        return response

    def query(
        self,
        query_text: str,
        /,
        parameters: dict[str, PythonParameter] | BaseModel | None = None,
        *,
        start_date: str | datetime.date,
        end_date: str | datetime.date,
        dest_table: str,
    ) -> QueryResult:
        return self._provide_response("query", QueryResult)

    @override
    def validate(
        self,
        query_text: str,
        /,
        parameters: dict[str, PythonParameter] | BaseModel | None = None,
    ) -> None:
        return self._provide_response("validate", NoneType)
